/*
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "cppheaderorganizer.h"
#include "cppheaderorganizer.moc"

#include "inclusionwizard.h"

#include "ast.h"
#include "cppsupportpart.h"
#include "backgroundparser.h"
#include "kdevpartcontroller.h"
#include "kdevdriver.h"
#include <kdevcoderepository.h>
#include "urlutil.h"
#include <catalog.h>
#include <codemodel.h>
#include <lookup.h>
#include <lexer.h>
#include <keywords.lut.h>

#include <ktexteditor/document.h>
#include <ktexteditor/editinterface.h>

#include <kdebug.h>

#include <qfileinfo.h>
#include <qdir.h>
#include <qregexp.h>


CppHeaderOrganizer::CppHeaderOrganizer(CppSupportPart* support)
  : QObject(support), support(support), editor(0)
{
  connect(support->partController(), SIGNAL(activePartChanged(KParts::Part*)),
	  SLOT(activePartChanged(KParts::Part*)));
//   connect(support, SIGNAL(fileParsed(const QString&)), 
// 	  SLOT(fileParsed(const QString&)));
}

void CppHeaderOrganizer::organizeHeaders()
{
  kdDebug() << "organizeHeaders called on " << activeFileName << endl;
  ids.clear();
  types.clear();
  forwardLine = -1;
  localIncludeLine = -1;
  globalIncludeLine = -1;
  firstASTLine = -1;
  markAlreadyIncludedTypesResolved();
  BackgroundParser* parser = support->backgroundParser();
  parser->lock();
  TranslationUnitAST* unit = parser->translationUnit(activeFileName);
  if (unit) {
    kdDebug() << " got a unit for " << activeFileName << endl;
    parseTranslationUnit(unit);
  }
  parser->unlock();

  QStringList localToInclude;
  QStringList globalToInclude;
  QStringList toForward;
  InclusionWizard wizard(0, true);
  for (QMap<QString, Candidate>::ConstIterator i(types.begin()); 
       i != types.end(); ++i) {
    if (i.data().type == Candidate::Include) {
      QStringList locals(getLocalHeadersForType(i.key()));
      QStringList globals(getGlobalHeadersForType(i.key()));
      if (locals.count() + globals.count() == 1) {
	localToInclude += locals;
	globalToInclude += globals;
      }
      else if (locals.count() + globals.count() > 1) {
	wizard.addConflict(i.key(), locals, globals);
      }
    }
    else if (i.data().type < Candidate::Include) {
      toForward << i.key();
    }
  }
  if (wizard.pageCount()) {
    if (wizard.exec() != QDialog::Accepted) {
      return;
    }
    localToInclude += wizard.getLocalResolvedIncludes();
    globalToInclude += wizard.getGlobalResolvedIncludes();
  }

  removeDuplicates(&localToInclude);
  removeDuplicates(&globalToInclude);
  subtract(&localToInclude, getIncludes(Local));
  subtract(&globalToInclude, getIncludes(Global));

  kdDebug() << "Final local files to include " << localToInclude << " at line " 
	    << localIncludeLine << endl;
  kdDebug() << "Final global files to include " << globalToInclude 
	    << " at line " << globalIncludeLine << endl;
  kdDebug() << "Final files to forward " << toForward << " at line "
	    << forwardLine << endl;
  insert(toForward, Candidate::Forward, Local);
  insert(globalToInclude, Candidate::Include, Global);
  insert(localToInclude, Candidate::Include, Local);
}

void CppHeaderOrganizer::activePartChanged(KParts::Part* part)
{
  editor = dynamic_cast<KTextEditor::EditInterface*>(part);
  if (KTextEditor::Document* doc = dynamic_cast<KTextEditor::Document*>(part)) {
    activeFileName = URLUtil::canonicalPath(doc->url().path());
  }
  else {
    activeFileName = QString::null;
  }
}

void CppHeaderOrganizer::fileParsed(const QString& file)
{
  activeFileName = file;
//   organizeHeaders();
}

void CppHeaderOrganizer::parseTranslationUnit(TranslationUnitAST* unit)
{
  TreeParser::parseTranslationUnit(unit);
}

void CppHeaderOrganizer::parseDeclaration(DeclarationAST* decl)
{
  TreeParser::parseDeclaration(decl);
}

void CppHeaderOrganizer::parseLinkageSpecification(LinkageSpecificationAST* link)
{
  kdDebug() << "LinkageSpecificationAST " << link->text() << endl;
  if (link->externType()) {
    kdDebug() << "LinkageSpecificationAST extern type " 
	      << link->externType()->text() << endl;
  }
  TreeParser::parseLinkageSpecification(link);
}

void CppHeaderOrganizer::parseNamespace( NamespaceAST* nspace)
{
  kdDebug() << "NamespaceAST " << nspace->text() << endl;
  if (nspace->namespaceName()) {
    kdDebug() << "NamespaceAST name " << nspace->namespaceName()->text() 
	      << endl;
  TreeParser::parseNamespace(nspace);
  }
}

void CppHeaderOrganizer::parseNamespaceAlias(NamespaceAliasAST* alias)
{
  // do we have that in c++ ?
  kdDebug() << "NamespaceAliasAST called" << alias->text() << endl;
}

void CppHeaderOrganizer::parseUsing( UsingAST* u)
{
  // TODO
  setFirstASTLine(u);
}

void CppHeaderOrganizer::parseUsingDirective( UsingDirectiveAST* u)
{
  // TODO
  setFirstASTLine(u);
}

void CppHeaderOrganizer::parseTypedef(TypedefAST* def)
{
  kdDebug() << "TypedefAST " << def->text() << endl;
  if (def->typeSpec()) {
    parseTypeSpecifier(def->typeSpec());
  }
  if (def->typeSpec() && def->initDeclaratorList()) {
    parseInitDeclaratorList(def->typeSpec(), def->initDeclaratorList());
  }
}

void CppHeaderOrganizer::parseTemplateDeclaration( TemplateDeclarationAST* t)
{
  // TODO
  setFirstASTLine(t);
}

void CppHeaderOrganizer::parseSimpleDeclaration( SimpleDeclarationAST* decl)
{
  kdDebug() << "SimpleDeclarationAST " << decl->text() << endl;
  if (decl->typeSpec()) {
    parseTypeSpecifier(decl->typeSpec());
  }
//    if (decl->functionSpecifier()) {
//      parseGroup(decl->functionSpecifier());
//    }
//   if (decl->storageSpecifier()) {
//     parseGroup(decl->storageSpecifier());
//   }
  if (decl->typeSpec() && decl->initDeclaratorList()) {
    parseInitDeclaratorList(decl->typeSpec(), decl->initDeclaratorList());
  }
  else {
    kdDebug() << " decl->initDeclaratorList() 0 " << endl;
  }
  if (decl->winDeclSpec()) {
    parseGroup(decl->winDeclSpec());
  }
}

// TODO create lookup tables local to the function
void CppHeaderOrganizer::parseFunctionDefinition(FunctionDefinitionAST* func)
{
  setFirstASTLine(func);
  kdDebug() << "FunctionDefinitionAST " << func->text() << endl;
  if (func->typeSpec()) {
    parseTypeSpecifier(func->typeSpec());
  }
//   if (func->functionSpecifier()) {
//     parseGroup(func->functionSpecifier());
//   }
//   if (func->storageSpecifier()) {
//     parseGroup(func->functionSpecifier());
//   }
  if (func->typeSpec() && func->initDeclarator()) {
    parseInitDeclarator(func->typeSpec(), func->initDeclarator());
  }
//   if (func->winDeclSpec()) {
//     parseGroup(func->winDeclSpec());
//   }
  if (func->functionBody()) {
    parseStatements(func->functionBody());
  }
}

void CppHeaderOrganizer::parseLinkageBody( LinkageBodyAST* )
{

}

void CppHeaderOrganizer::parseAccessDeclaration( AccessDeclarationAST* decl)
{
  kdDebug() << "AccessDeclarationAST " << decl->text() << endl;
}

void CppHeaderOrganizer::parseTypeSpecifier( TypeSpecifierAST* type)
{
  setFirstASTLine(type);
  kdDebug() << "TypeSpecifierAST " << type->text() << endl;
  if (type->cvQualify()) {
    parseGroup(type->cvQualify());
  }
  if (type->cv2Qualify()) {
    parseGroup(type->cv2Qualify());
  }
  TreeParser::parseTypeSpecifier(type);
}

/**
 * For a class specifier we have to parse its base clause to make sure the
 * header files of all its parents are included.
 */
void CppHeaderOrganizer::parseClassSpecifier(ClassSpecifierAST* klass)
{
  kdDebug() << "ClassSpecifierAST " << klass->text() << endl;
  if (klass->baseClause()) {
    parseBaseClause(klass->baseClause());
  }
  // TreeParser handles parsing of declarations inside the class
  TreeParser::parseClassSpecifier(klass);
}

void CppHeaderOrganizer::parseEnumSpecifier( EnumSpecifierAST* )
{

}

void CppHeaderOrganizer::parseElaboratedTypeSpecifier
(ElaboratedTypeSpecifierAST* type)
{
  // hope name is always != 0
  if (type->name()) {
    QString name(type->name()->text());
    kdDebug() << "ElaboratedTypeSpecifierAST " << name << endl;
    // mark type as resolved
    markType(name, Candidate::Forwarded);
    type->getEndPosition(&forwardLine, 0);
  }
}

void CppHeaderOrganizer::parseName(NameAST* name)
{
  kdDebug() << "NameAST " << name->text() << endl;
}

void CppHeaderOrganizer::parseStatements(StatementListAST* l)
{
  kdDebug() << "statement list: new block?? " << endl << endl << endl;
  QPtrList<StatementAST> list(l->statementList());
  for (QPtrListIterator<StatementAST> i(list); i.current(); ++i) {
    parseStatement(i.current());
  }
}

void CppHeaderOrganizer::parseStatement(StatementAST* stat)
{
  if (!stat) {
    return;
  }

  kdDebug() << "StatementAST " << stat->text() << endl;

  switch (stat->nodeType()) {
  case NodeType_DeclarationStatement:
    parseDeclaration
      (static_cast<DeclarationStatementAST*>(stat)->declaration());
    break;
  case NodeType_DoStatement:
    parseCondition(static_cast<DoStatementAST*>(stat)->condition());
    parseStatement(static_cast<DoStatementAST*>(stat)->statement());
    break;
  case NodeType_ExpressionStatement:
    parseExpressionStatement(static_cast<ExpressionStatementAST*>(stat));
    break;
  case NodeType_ForStatement:
    parseStatement(static_cast<ForStatementAST*>(stat)->initStatement());
    parseCondition(static_cast<ForStatementAST*>(stat)->condition());
    parseExpression(static_cast<ForStatementAST*>(stat)->expression());
    parseStatement(static_cast<ForStatementAST*>(stat)->statement());
    break;
  case NodeType_IfStatement:
    parseCondition(static_cast<IfStatementAST*>(stat)->condition());
    parseStatement(static_cast<IfStatementAST*>(stat)->statement());
    parseStatement(static_cast<IfStatementAST*>(stat)->elseStatement());
    break;
  case NodeType_SwitchStatement:
    parseCondition(static_cast<SwitchStatementAST*>(stat)->condition());
    parseStatement(static_cast<SwitchStatementAST*>(stat)->statement());
    break;
  case NodeType_WhileStatement:
    parseCondition(static_cast<WhileStatementAST*>(stat)->condition());
    parseStatement(static_cast<WhileStatementAST*>(stat)->statement());
    break;
  case NodeType_StatementList:
    parseStatements(static_cast<StatementListAST*>(stat));
    break;
  default:
    kdDebug() << "unhandled statement " <<  stat->nodeType() << " " 
	      << nodeTypeToString(stat->nodeType()) << endl;
  }
}

void CppHeaderOrganizer::parseGroup(GroupAST* group)
{
  kdDebug() << "GroupAST " << group->text() << endl;
}

void CppHeaderOrganizer::parseCondition(ConditionAST* cond)
{
  if (!cond) {
    return;
  }
  kdDebug() << "ConditionAST " << cond->text() << endl;

//   if (cond->typeSpec()) {
//     parseTypeSpecifier(cond->typeSpec());
//   }
//   if (cond->declarator()) {
//     parseDeclarator(cond->declarator());
//   }
  if (cond->declarator()) {
    if (cond->typeSpec()) {
      extract(cond->typeSpec(), cond->declarator());
    }
    parseDeclarator(cond->declarator());
  }
  if (cond->expression()) {
    parseExpression(cond->expression());
  }
}

void CppHeaderOrganizer::parseDeclarator(DeclaratorAST* decl)
{
  kdDebug() << "DeclaratorAST " << decl->text() << endl;
  if (decl->subDeclarator()) {
    // TODO what is a sub declarator?
    parseDeclarator(decl->subDeclarator());
  }
  // if (decl->declaratorId()) {
//     parseName(decl->declaratorId());
//   }
  if (decl->parameterDeclarationClause()) {
    parseParameterDeclarationClause(decl->parameterDeclarationClause());
  }
  // TODO exceptions
}

void CppHeaderOrganizer::parseInitDeclaratorList(TypeSpecifierAST* type, 
						 InitDeclaratorListAST* l)
{
  kdDebug() << "InitDeclaratorListAST " << l->text() << endl;
  QPtrList<InitDeclaratorAST> list(l->initDeclaratorList());
  for (QPtrListIterator<InitDeclaratorAST> i(list); i.current(); ++i) {
    parseInitDeclarator(type, i.current());
  }
}

void CppHeaderOrganizer::parseInitDeclarator(TypeSpecifierAST* type, 
					     InitDeclaratorAST* init)
{
  kdDebug() << "InitDeclaratorAST " << init->text() <<  "for type " 
	    << type->text() << endl;
  if (init->declarator()) {
    extract(type, init->declarator());
    parseDeclarator(init->declarator());
  }
  // TODO initializer
  if (init->initializer()) {
    kdDebug() << "InitDeclaratorAST initializer " << init->initializer()->text()
	      << endl;
    parseExpression(init->initializer());
  }
}

/**
 * Parse specifiers.
 */
void CppHeaderOrganizer::parseBaseClause(BaseClauseAST* base)
{
  kdDebug() << "BaseClauseAST " << base->text() << endl;
  QPtrList<BaseSpecifierAST> list(base->baseSpecifierList());
  for (QPtrListIterator<BaseSpecifierAST> i(list); i.current(); ++i) {
    parseBaseSpecifier(i.current());
  }
}

/**
 * For a base class specifier we are not interested in its attributes
 * (virtual, public, protected etc.) just in its name.
 */
void CppHeaderOrganizer::parseBaseSpecifier(BaseSpecifierAST* spec)
{
  kdDebug() << "BaseSpecifierAST " << spec->text() << endl;
  if (spec->name()) {
    typeNeedsInclude(spec->name()->text());
  }
}

void CppHeaderOrganizer::parseParameterDeclarationClause
(ParameterDeclarationClauseAST* params)
{
  kdDebug() << "ParameterDeclarationClauseAST " << params->text() << endl;

  if (params->parameterDeclarationList()) {
    parseParameterDeclarationList(params->parameterDeclarationList());
  }
}

void CppHeaderOrganizer::parseParameterDeclarationList
(ParameterDeclarationListAST* paramlist)
{
  kdDebug() << "ParameterDeclarationListAST " << paramlist->text() << endl;
  QPtrList<ParameterDeclarationAST> list(paramlist->parameterList());
  for (QPtrListIterator<ParameterDeclarationAST> i(list); i.current(); ++i) {
    parseParameterDeclaration(i.current());
  }
}

void CppHeaderOrganizer::parseParameterDeclaration(ParameterDeclarationAST* decl)
{
  kdDebug() << "ParameterDeclarationAST " << decl->text() << endl;
  if (decl->declarator()) {
    if (decl->typeSpec()) {
      extract(decl->typeSpec(), decl->declarator());
    }
    parseDeclarator(decl->declarator());
  }
  if (decl->expression()) {
    parseExpression(decl->expression());
  }
}

void CppHeaderOrganizer::parseExpressionStatement(ExpressionStatementAST* exp)
{
  kdDebug() << "ExpressionStatementAST " << exp->text() << endl;
  if (exp->expression()) {
    parseExpression(exp->expression());
  }
}

// TODO parse all occurrences of . , ->, :: etc.
// TODO extract called functions and stack constructors
void CppHeaderOrganizer::parseExpression(AST* exp)
{
  if (!exp) {
    return;
  }
  kdDebug() << "Expression " << exp->text() << endl;
  QStringList tokens(QStringList::split(' ', exp->text()));
  QStringList::Iterator i;
  for (i = tokens.find("."); i != tokens.end(); i = tokens.find(++i, ".")) {
    if (i != tokens.begin()) {
      if(*(--i) != ")") {
	idNeedsInclude(*i);
      }
      ++i;
    }
  }
  for (i = tokens.find("->"); i != tokens.end(); i = tokens.find(++i, "->")){
    if (i != tokens.begin()) {
      if (*(--i) != ")") {
	idNeedsInclude(*i);
      }
      ++i;
    }
  }
  for (i = tokens.find("new"); i != tokens.end(); i = tokens.find(++i, "new")) {
    if (++i != tokens.end()) {
      typeNeedsInclude(*i);
    }
  }
  for(i = tokens.find("("); i != tokens.end(); i = tokens.find(++i, "(")) {
    if (i != tokens.begin()) {
      if (*(--i) != ")") {
	// typeNeedsInclude(*i);
      }
      ++i;
    }
  }
  for (i = tokens.find("::"); i != tokens.end(); i = tokens.find(++i, "::")) {
    if (i != tokens.begin()) {
      typeNeedsInclude(*(--i));
      ++i;
    }
    // maybe add hint that this could be a namespace too
  }
}

void CppHeaderOrganizer::extract(TypeSpecifierAST* type, DeclaratorAST* decl)
{
  kdDebug() << "extract for type " << type->text() << " and decl " 
	    << decl->text() << endl;
  if (type->name()) {
    QString typeName(type->name()->text());
    if (!decl->ptrOpList().isEmpty()) {
      typeNeedsForward(typeName);
    }
    else {
      typeNeedsInclude(typeName);
    }
    if (decl->declaratorId()) {
      ids[decl->declaratorId()->text()] = typeName;
    }
  }
}

void CppHeaderOrganizer::idNeedsInclude(const QString& id)
{
  kdDebug() << "id needs include " << id << endl;
  QMap<QString, QString>::const_iterator i(ids.find(id));
  if (i != ids.end()) {
    typeNeedsInclude(*i);
  }
  else {
    // TODO look into header, or just ignore, could be protected member from
    // super class
  }
}

void CppHeaderOrganizer::typeNeedsInclude(const QString& type)
{
  kdDebug() << "type needs include " << type << endl;
  markType(type, Candidate::Include);
}

void CppHeaderOrganizer::typeNeedsForward(const QString& type)
{
  kdDebug() << "type needs forward " << type << endl;
  markType(type, Candidate::Forward);
}

void CppHeaderOrganizer::markType(const QString& type, Candidate::Type t)
{
  if (isPrimitive(type)) {
    return;
  }

  QMap<QString, Candidate>::iterator i(types.find(type));

  if (i != types.end()) {
    if (i.data().type < t) {
      i.data().type = t;
    }
  }
  else {
    types[type].type = t;
  }
}

bool CppHeaderOrganizer::isPrimitive(const QString& type)
{
  QStringList subtypes(QStringList::split(' ', type));
  const QStringList::ConstIterator end(subtypes.end());
  for (QStringList::ConstIterator i(subtypes.begin()); i != end; ++i) {
    if (Lookup::findEntry(&keyword, *i)) {
      return true;
    }
  }
  return false;
}

QStringList CppHeaderOrganizer::getLocalHeadersForType(const QString& type)
{
  QStringList results;
  
  // local lookup in codemodel
  const FileList files(support->codeModel()->fileList());
  for (FileList::ConstIterator i(files.begin()); i != files.end(); ++i) {
    const FileDom f(*i);
    if ((*f).hasClass(type) || (*f).hasFunction(type) || (*f).hasVariable(type)
	|| (*f).hasEnum(type) || (*f).hasTypeAlias(type)) {
      kdDebug() << "Found local type " << type << " in " << (*f).name() << endl;
      // just use filename for local files
      // results << "\"" + QFileInfo((*f).name()).fileName() + "\"";
      results << QFileInfo((*f).name()).fileName();
    }
  }
  return results;
}

QStringList CppHeaderOrganizer::getGlobalHeadersForType(const QString& type)
{
  QStringList results;

  // global lookup in catalogs
  QValueList<Catalog*> catalogs(support->codeRepository()->registeredCatalogs());
  QValueList<Catalog::QueryArgument> args;
  args << Catalog::QueryArgument("name", type);

  for (QValueList<Catalog*>::ConstIterator it(catalogs.begin()); 
       it!= catalogs.end(); ++it) {
    Catalog* catalog = *it;
    if(catalog->enabled()) {
      QValueList<Tag> tags(catalog->query(args));
      if (!tags.isEmpty()) {
	QStringList paths(support->getIncludePaths(catalog));
  	results += stripPrefixes(tags, paths);
      }
    }
  }
  removeDuplicates(&results);
  return results;
}

QStringList CppHeaderOrganizer::stripPrefixes(const QValueList<Tag>& tags, 
					      const QStringList& paths)
{
  QStringList results;
  // TODO cache everything
  for (QValueList<Tag>::ConstIterator i(tags.begin()); i != tags.end(); ++i) {
    for (QStringList::ConstIterator p(paths.begin()); p != paths.end(); ++p) {
      bool isParent = false;
      QString stripped(KURL::relativePath(*p, (*i).fileName(), &isParent));
      if (isParent) {
	// remove "./"
	results << stripped.right(stripped.length() - 2);
	break;
      }
    }
  }
  return results;
}

/**
 * Removes duplicate entries to files having the same file name but are in
 * different directories.
 */
void CppHeaderOrganizer::removeDuplicates(QStringList* files)
{
  bool seen = true;
  QDict<bool> duplicates;
  QStringList::Iterator i(files->begin());
  while (i != files->end()) {
    if (duplicates.find(*i)) {
      i = files->remove(i);
    }
    else {
      duplicates.insert(*i, &seen);
      ++i;
    }
  }
}

void CppHeaderOrganizer::subtract(QStringList* includes, 
				  const QStringList& already)
{
  QStringList::Iterator i(includes->begin());
  while(i != includes->end()) {
    if (already.findIndex(*i) != -1) {
      i = includes->remove(i);
    }
    else {
      ++i;
    }
  }
}

QStringList CppHeaderOrganizer::getAllIncludes()
{
  QStringList results(getIncludes(Local));
  results += getIncludes(Global);
  return results;
}

QStringList CppHeaderOrganizer::getIncludes(IncludeType type)
{
  if (!editor) {
    return QStringList();
  }

  QString text(editor->text());

  QStringList results;
  QRegExp exp = (type == Global) ? QRegExp("#include\\s*<([^>\n]+)>")
    : QRegExp("#include\\s*\"([^\"\n]+)\"");
  int pos = 0;
  int lastpos = -1;
  while ((pos = exp.search(text, pos)) != -1) {
    results << exp.cap(1);
    pos += exp.matchedLength();
    lastpos = exp.pos(0);
  }
  setLastIncludeLine(lastpos, type);
  return results;
}

void CppHeaderOrganizer::setLastIncludeLine(int pos, IncludeType type)
{
  kdDebug() << "setLastIncludeLine for " << pos << " and type " << type << endl;
  int& line = (type == Local) ? localIncludeLine : globalIncludeLine;
  // only compute this once
  if (pos == -1 || line != -1 || !editor) {
    return;
  }
  uint lines = editor->numLines();
  for (uint i = 0; i < lines; i++) {
    if (pos > 0) {
      // + 1 = new line character
      pos -= (editor->lineLength(i) + 1);
    }
    else {
      if (i < lines - 1) {
	line = ++i;
      }
      else {
	line = 0;
      }
      break;
    }
  }
}

QStringList CppHeaderOrganizer::computeMissingIncludes
(const QStringList& existingIncludes, const QStringList& newIncludes)
{
  bool included = true;
  QDict<bool> existing;
  for (QStringList::ConstIterator i(existingIncludes.begin()); 
       i != existingIncludes.end(); ++i) {
    existing.insert(*i, &included);
  }
  QStringList results;
  for (QStringList::ConstIterator i(newIncludes.begin()); 
       i != newIncludes.end(); ++i) {
    if (!existing.find(*i)) {
      results << *i;
    }
  }
  return results;
}

void CppHeaderOrganizer::markAlreadyIncludedTypesResolved()
{
  QStringList local(getIncludes(Local));
  const FileList files(support->codeModel()->fileList());
  for (QStringList::ConstIterator i(local.begin()); i != local.end(); ++i) {
    for (FileList::ConstIterator j(files.begin()); j != files.end(); ++j) {
      const FileDom f(*j);
      kdDebug() << "file dom " << (*f).name() << " name " << *i << endl;
      if ((*f).name().endsWith(*i)) {
	const ClassList classes((*f).classList());
	for (ClassList::ConstIterator c(classes.begin()); c != classes.end();
	     ++c) {
	  const ClassDom clazz(*c);
	  kdDebug() << "marking class "  << (*clazz).name() << " as resolved" 
		    << endl;
	  markType((*clazz).name(), Candidate::Resolved);
	}
	// break here?
      }
    }
  }


  QStringList globals(getIncludes(Global));
  QValueList<Catalog*> catalogs(support->codeRepository()->registeredCatalogs());
  for (QValueList<Catalog*>::ConstIterator it(catalogs.begin()); 
       it!= catalogs.end(); ++it) {
    Catalog* catalog = *it;
    if(!catalog->enabled()) {
      continue;
    }
    QStringList prefixes(support->getIncludePaths(catalog));
    
    for (QStringList::ConstIterator i(globals.begin()); i != globals.end();
	 ++i) {
      
      QStringList fileNames(addPrefixes(prefixes, *i));
      
      for (QStringList::ConstIterator j(fileNames.begin()); 
	   j != fileNames.end(); ++j) {
	QValueList<Catalog::QueryArgument> args;
	args << Catalog::QueryArgument("fileName", *j);
        
	QValueList<Tag> tags(catalog->query(args));
	for (QValueList<Tag>::ConstIterator t(tags.begin()); t != tags.end();
	     ++t) {
	  // TODO add other tags too
	  if ((*t).kind() == Tag::Kind_Class) {
	    kdDebug() << "found class tag " << (*t).name() << endl;
	    markType((*t).name(), Candidate::Resolved);
	  }
	}
      }
    }
  }
}

QStringList CppHeaderOrganizer::addPrefixes(const QStringList& prefixes,
					    const QString& fileName)
{
  QStringList results;
  for (QStringList::ConstIterator i(prefixes.begin()); i != prefixes.end();
       ++i) {
    QDir dir(*i);
    QFileInfo info(dir, fileName);
    // we could check existence here too, if it's faster
    results << info.absFilePath();
  }
  return results;
}

void CppHeaderOrganizer::insert(const QStringList& types, Candidate::Type type, 
				IncludeType iType)
{
  if (!editor) {
    return;
  }
  // TODO update other lines respectively
  if (type == Candidate::Forward) {
    int line = QMAX(0, (forwardLine != -1) ? forwardLine : firstASTLine - 1);
    for (QStringList::ConstIterator i(types.begin()); i != types.end(); ++i) {
      editor->insertLine(line++, "class " + *i + ";");
    }
  }
  else if (type == Candidate::Include) {
    if (iType == Global) {
      int line = QMAX(0, (globalIncludeLine != -1) ? globalIncludeLine
		      : firstASTLine - 1);
      for (QStringList::ConstIterator i(types.begin()); i != types.end(); ++i) {
	editor->insertLine(line++, "#include <" + *i + ">");
      }
    }
    else {
      int line = QMAX(0, (localIncludeLine != -1) ? localIncludeLine
		      : firstASTLine - 1);
      for (QStringList::ConstIterator i(types.begin()); i != types.end(); ++i) {
	editor->insertLine(line++, "#include \"" + *i + "\"");
      }
    }
  }
}

inline void CppHeaderOrganizer::setFirstASTLine(AST* ast)
{
  if (firstASTLine == -1) {
    ast->getStartPosition(&firstASTLine, 0);
  }
  else {
    int line;
    ast->getStartPosition(&line, 0);
    firstASTLine = QMIN(line, firstASTLine);
  }
}

/*
 * Local Variables:
 * compile-command: "make install-kde_moduleLTLIBRARIES"
 * End:
 */
