/*
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef CPPHEADERORGANIZER_H
#define CPPHEADERORGANIZER_H

#include <qobject.h>
#include <tree_parser.h>
#include <catalog.h>

#include <qmap.h>

class CppSupportPart;

namespace KParts 
{
  class Part;
}

namespace KTextEditor
{
  class EditInterface;
}

class CppHeaderOrganizer : public QObject, public TreeParser
{
  Q_OBJECT

public:
  CppHeaderOrganizer(CppSupportPart* support);

  // translation-unit
  virtual void parseTranslationUnit( TranslationUnitAST* );
    
  // declarations
  virtual void parseDeclaration( DeclarationAST* );
  virtual void parseLinkageSpecification( LinkageSpecificationAST* );
  virtual void parseNamespace( NamespaceAST* );
  virtual void parseNamespaceAlias( NamespaceAliasAST* );
  virtual void parseUsing( UsingAST* );
  virtual void parseUsingDirective( UsingDirectiveAST* );
  virtual void parseTypedef( TypedefAST* );
  virtual void parseTemplateDeclaration( TemplateDeclarationAST* );
  virtual void parseSimpleDeclaration( SimpleDeclarationAST* );
  virtual void parseFunctionDefinition( FunctionDefinitionAST* );
  virtual void parseLinkageBody( LinkageBodyAST* );
  virtual void parseAccessDeclaration( AccessDeclarationAST* );

  // type-specifier
  virtual void parseTypeSpecifier( TypeSpecifierAST* );
  virtual void parseClassSpecifier( ClassSpecifierAST* );
  virtual void parseEnumSpecifier( EnumSpecifierAST* );
  virtual void parseElaboratedTypeSpecifier( ElaboratedTypeSpecifierAST* );

  virtual void parseName(NameAST*);
  virtual void parseStatements(StatementListAST*);
  virtual void parseStatement(StatementAST*);
  virtual void parseGroup(GroupAST*);
  virtual void parseCondition(ConditionAST*);
  virtual void parseDeclarator(DeclaratorAST*);
  virtual void parseInitDeclaratorList(TypeSpecifierAST*, 
				       InitDeclaratorListAST*);
  virtual void parseInitDeclarator(TypeSpecifierAST*, InitDeclaratorAST*);
  virtual void parseBaseClause(BaseClauseAST*);
  virtual void parseBaseSpecifier(BaseSpecifierAST*);
  virtual void parseParameterDeclarationClause(ParameterDeclarationClauseAST*);
  virtual void parseParameterDeclarationList(ParameterDeclarationListAST*);
  virtual void parseParameterDeclaration(ParameterDeclarationAST*);
  virtual void parseExpressionStatement(ExpressionStatementAST*);
  virtual void parseExpression(AST*);

  void extract(TypeSpecifierAST*, DeclaratorAST*);
  void idNeedsInclude(const QString& id);
  void typeNeedsInclude(const QString& type);
  void typeNeedsForward(const QString& type);
  bool isPrimitive(const QString& type);
  QStringList getLocalHeadersForType(const QString& type);
  QStringList getGlobalHeadersForType(const QString& type);
  void removeDuplicates(QStringList* files);
  void subtract(QStringList* includes, const QStringList& already);
  QStringList getAllIncludes();
  QStringList computeMissingIncludes(const QStringList&, const QStringList&);
  QStringList stripPrefixes(const QValueList<Tag>& tags, 
			    const QStringList& paths);
  QStringList addPrefixes(const QStringList& prefixes, const QString& fileName);
  void markAlreadyIncludedTypesResolved();


  enum IncludeType {
    Global, Local
  };

  QStringList getIncludes(IncludeType type);
  void setLastIncludeLine(int pos, IncludeType type);
  inline void setFirstASTLine(AST* ast);

  public slots:
    void organizeHeaders();

  private slots:
    void activePartChanged(KParts::Part* part);
    void fileParsed(const QString& file);

 private:
  CppSupportPart* support;
  KTextEditor::EditInterface* editor;
  int forwardLine;
  int localIncludeLine;
  int globalIncludeLine;
  int firstASTLine;
  QString activeFileName;
  
  struct Candidate
  {
    Candidate()
      : type(0)
    {
    }
    QString name;
    enum Type {
      Forward = 1, Extern, Forwarded, Include, Resolved
    };
    int type;
  };

  void markType(const QString&, Candidate::Type);
  void insert(const QStringList& types, Candidate::Type type, 
	      IncludeType iType);
  
  QMap<QString, Candidate> types;
  QMap<QString, QString> ids;
};

#endif
