#include "inclusionwizard.h"
#include "inclusionwizard.moc"

#include <klocale.h>
#include <kdebug.h>
#include <klistviewsearchline.h> 

#include <qheader.h>
#include <qpushbutton.h>

InclusionWizard::InclusionWizard(QWidget* parent, bool modal)
  : KWizard(parent, "inclusion wizard", modal)
{
  
}

void InclusionWizard::addConflict(const QString& type, 
				  const QStringList& locals,
				  const QStringList& globals)
{
  InclusionPage* page = new InclusionPage(this, type, locals, globals);
  addPage(page, i18n("Select proper header file for \"%1\"").arg(type));
}

QStringList InclusionWizard::getLocalResolvedIncludes() const
{
  QStringList results;
  for (int i = 0; i < pageCount(); i++) {
    InclusionPage* p = static_cast<InclusionPage*>(page(i));
    if (p->isLocal()) {
      results << p->getResolved();
    }
  }
  return results;
}

QStringList InclusionWizard::getGlobalResolvedIncludes() const
{
  QStringList results;
  for (int i = 0; i < pageCount(); i++) {
    InclusionPage* p = static_cast<InclusionPage*>(page(i));
    if (!p->isLocal()) {
      results << p->getResolved();
    }
  }
  return results;
}

InclusionSearchLine::InclusionSearchLine(QWidget* parent, const char* name)
  : KListViewSearchLine(parent, name)
{
}

void InclusionSearchLine::updateSearch(const QString& s)
{
  KListViewSearchLine::updateSearch(s);
  emit updatedSearch();
}

void InclusionSearchLine::keyPressEvent(QKeyEvent* event)
{
  if (event->key() == Qt::Key_Down) {
    event->accept();
    focusNextPrevChild(true);
    listView()->setSelected(listView()->firstChild(), true);
  }
  else {
    KListViewSearchLine::keyPressEvent(event);
  }
}

InclusionPage::InclusionPage(InclusionWizard* parent, const QString& /*type*/,
			     const QStringList& locals, 
			     const QStringList& globals)
  : QVBox(parent), wizard(parent), local(false)
{
  search = new InclusionSearchLine(this, "search line");
  connect(search, SIGNAL(updatedSearch()), SLOT(verifyInput()));
  connect(search, SIGNAL(returnPressed(const QString&)), SLOT(executed()));

  listView = new HeaderItemListView(this);
  search->setListView(listView);
  listView->header()->hide();
  listView->header()->setStretchEnabled(true, 0);
  listView->addColumn(i18n("Header"));
  listView->setSorting(0);
  connect(listView, SIGNAL(selectionChanged(QListViewItem*)), 
	  SLOT(selectionChanged(QListViewItem*)));
  connect(listView, SIGNAL(executed(QListViewItem*)), SLOT(executed()));
  for (QStringList::ConstIterator i(locals.begin()); i != locals.end(); ++i) {
    new HeaderItem(listView, *i, true);
  }
  for (QStringList::ConstIterator i(globals.begin()); i != globals.end(); ++i) {
    new HeaderItem(listView, *i, false);
  }
}

void InclusionPage::showEvent(QShowEvent* event)
{
  if (visibleItemsCount() <= 1 || listView->selectedItem()) {
    setNextOrFinished(true);
  }
  else {
    setNextOrFinished(false);
  }
  search->setFocus();
  QVBox::showEvent(event);
}

bool InclusionPage::setNextOrFinished(bool enable)
{
  if (wizard->indexOf(this) == wizard->pageCount() - 1) {
    wizard->setFinishEnabled(this, enable);
    return true;
  }
  else {
    wizard->setNextEnabled(this, enable);
    return false;
  }
}

QString InclusionPage::getResolved() const
{
  return file;
}

bool InclusionPage::isLocal() const
{
  return local;
}

void InclusionPage::verifyInput()
{
  int count = visibleItemsCount();
  if (count == 1) {
    QListViewItemIterator i(listView, QListViewItemIterator::Visible);
    if (HeaderItem* h = static_cast<HeaderItem*>(i.current())) {
      file = h->getFile();
      local = h->isLocal();
      setNextOrFinished(true);
    }
  }
  else if (count == 0) {
    file = search->text();
    setNextOrFinished(true);
  }
  else {
    setNextOrFinished(false);
  }
}

void InclusionPage::executed()
{
  if (wizard->finishButton()->isEnabled() 
      && wizard->finishButton()->isVisible()) {
    wizard->accept();
  }
  else if (wizard->nextButton()->isEnabled()
	   && wizard->nextButton()->isVisible()) {
    wizard->next();
  }
}

void InclusionPage::selectionChanged(QListViewItem* item)
{
  if (HeaderItem* h = static_cast<HeaderItem*>(item)) {
    file = h->getFile();
    local = h->isLocal();
    setNextOrFinished(true);
  }
  else {
    setNextOrFinished(false);
  }
}

int InclusionPage::visibleItemsCount()
{
  int count = 0;
  for (QListViewItemIterator i(listView, QListViewItemIterator::Visible);
       i.current(); ++i, ++count);
  return count;
}

HeaderItem::HeaderItem(QListView* parent, const QString& file, bool local)
  : KListViewItem(parent, local ? "\"" + file + "\"" : "<" + file + ">"),
    file(file), local(local)
{
}

const QString& HeaderItem::getFile() const
{
  return file;
}

bool HeaderItem::isLocal() const
{
  return local;
}

HeaderItemListView::HeaderItemListView(QWidget* parent)
  : KListView(parent)
{
}

void HeaderItemListView::keyPressEvent(QKeyEvent* event)
{
  if (event->key() == Qt::Key_Return || event->key() == Qt::Key_Enter) {
    if (selectedItem()) {
      emit executed(selectedItem());
      return;
    }
  }
  KListView::keyPressEvent(event);
}


/*
 * Local Variables:
 * compile-command: "make install-kde_moduleLTLIBRARIES"
 * End:
 */
